// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef READ_CONFIG_HXX
#define READ_CONFIG_HXX

struct Config;

void
file_read_config(Config &config);

#endif
